c down
alt down
` down
` up
tab down
tab up
tab down
tab up
c up
a down
a up
x down
x up
alt up
c up

control down
alt down
alt up
shift down
x down
x up
shift up
shift down
x down
x up
shift up
control up
b down
b up
x down
x up
