alt down
` down
` up
` down
` up
tab down
tab up
tab down
tab up
a down
a up
alt up

alt down
control down
alt up
control up
9 down
9 up
shift down
x down
x up
shift up
shift down
x down
x up
shift up
b down
b up
