meta down
a down
a up
meta up
rightmeta down
a down
a up
rightmeta up

meta down
control down
meta up
control up
b down
b up
meta down
control down
meta up
control up
meta down
control down
meta up
control up
b down
b up
meta down
control down
meta up
control up
