alt down
1 down
1 up
tab down
tab up
tab down
tab up
a down
a up
x down
x up
alt up

alt down
control down
alt up
control up
tab down
tab up
shift down
x down
x up
shift up
shift down
x down
x up
shift up
b down
b up
x down
x up
